"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CustomizedLanguageService = void 0;
const constants_1 = require("./constants");
const types_1 = require("./types");
const utils_1 = require("./utils");
const helper_1 = require("./helper");
class CustomizedLanguageService {
    constructor(info, typescript, logger) {
        this.info = info;
        this.typescript = typescript;
        this.logger = logger;
    }
    getApplicableRefactors(fileName, positionOrRange, preferences, triggerReason) {
        const [startPosition, endPosition] = utils_1.getRangeOfPositionOrRange(positionOrRange);
        if (!helper_1.isDef(endPosition)) {
            this.logger.log('Cannot find endPosition');
            return [];
        }
        const context = this.getRefactorContext(fileName);
        if (!context) {
            this.logger.log('Cannot refactor context');
            return [];
        }
        const { file, program } = context;
        const info = this.getInfo(startPosition, endPosition, file, program, false);
        if (!info) {
            this.logger.log('Empty info');
            return [];
        }
        if (info.kind === types_1.RefactorKind.useCallback) {
            return [
                {
                    name: constants_1.refactorName,
                    description: constants_1.refactorDescriptions,
                    actions: [
                        {
                            name: constants_1.wrapIntoUseCallbackActionName,
                            description: constants_1.wrapIntoUseCallbackActionDescription
                        }
                    ]
                }
            ];
        }
        if (info.kind === types_1.RefactorKind.useMemo) {
            return [
                {
                    name: constants_1.refactorName,
                    description: constants_1.refactorDescriptions,
                    actions: [
                        {
                            name: constants_1.wrapIntoUseMemoActionName,
                            description: constants_1.wrapIntoUseMemoActionDescription
                        }
                    ]
                }
            ];
        }
        return [];
    }
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        const [startPosition, endPosition] = utils_1.getRangeOfPositionOrRange(positionOrRange);
        if (!helper_1.isDef(endPosition)) {
            this.logger.log('Cannot find endPosition');
            return undefined;
        }
        const context = this.getRefactorContext(fileName);
        if (!context) {
            this.logger.log('Cannot refactor context');
            return undefined;
        }
        const { file, program } = context;
        const info = this.getInfo(startPosition, endPosition, file, program, true);
        if (!info) {
            this.logger.log('Empty info');
            return undefined;
        }
        const formatContext = this.typescript.formatting.getFormatContext(formatOptions);
        const textChangesContext = {
            formatContext,
            host: this.info.languageServiceHost,
            preferences: preferences || {}
        };
        if (info.kind === types_1.RefactorKind.useCallback &&
            actionName === constants_1.wrapIntoUseCallbackActionName) {
            return this.getEditsForConvertUseCallback(info, file, textChangesContext);
        }
        if (info.kind === types_1.RefactorKind.useMemo &&
            actionName === constants_1.wrapIntoUseMemoActionName) {
            return this.getEditsForConvertUseMemo(info, file, textChangesContext);
        }
        this.logger.log('Unknown action');
        return undefined;
    }
    getInfo(startPosition, endPosition, file, program, full) {
        const ts = this.typescript;
        const startToken = utils_1.skipJsxTextToken(ts, ts.getTokenAtPosition(file, startPosition), file);
        const rawTopLevelNode = utils_1.findTopLevelNodeInSelection(ts, startToken, startPosition, endPosition, file);
        if (!rawTopLevelNode) {
            this.logger.log('Cannot find rawTopLevelNode');
            return undefined;
        }
        const topLevelNode = utils_1.skipTriviaExpression(this.typescript, utils_1.skipJsxExpression(this.typescript, utils_1.skipSingleValueDeclaration(this.typescript, rawTopLevelNode)));
        if (utils_1.isDefinitelyNotSupportedToken(ts, topLevelNode)) {
            this.logger.log('topLevelNode definitely not supported: ' + topLevelNode.kind);
            return undefined;
        }
        if (ts.isPartOfTypeQuery(topLevelNode)) {
            this.logger.log('Cannot work inside type query');
            return undefined;
        }
        const checker = program.getTypeChecker();
        if (utils_1.alreadyWrappedOrContainsInReactHooks(ts, topLevelNode, checker)) {
            this.logger.log('Already has hooks');
            return undefined;
        }
        if (!utils_1.isInFunctionComponent(ts, topLevelNode, checker)) {
            this.logger.log('Not in function component');
            return undefined;
        }
        this.logger.log('TopLevelKind: ' + topLevelNode.kind);
        if (utils_1.isFunctionExpressionLike(ts, topLevelNode)) {
            return this.getInfoFromFunctionExpressionLike(topLevelNode, file, checker, full);
        }
        if (utils_1.isExpression(ts, topLevelNode)) {
            return this.getInfoFromUniversalExpression(topLevelNode, file, checker, full);
        }
        this.logger.log('No actions');
        return undefined;
    }
    getInfoFromUniversalExpression(expression, file, checker, full) {
        const deps = full
            ? this.getOutsideReferences(expression, [], file, checker)
            : [];
        const hooksReference = full
            ? utils_1.getHooksNameReferenceType(this.typescript, expression, checker, 'useMemo', this.isNewJsxTransformer())
            : undefined;
        this.logger.log('Universal Deps: ' + deps.length);
        this.logger.log('hooksReference: ' + JSON.stringify(hooksReference));
        return {
            kind: types_1.RefactorKind.useMemo,
            expression,
            deps,
            hooksReference
        };
    }
    getInfoFromFunctionExpressionLike(func, file, checker, full) {
        const deps = full
            ? this.getOutsideReferences(func.body, func.parameters, file, checker)
            : [];
        const hooksReference = full
            ? utils_1.getHooksNameReferenceType(this.typescript, func, checker, 'useCallback', this.isNewJsxTransformer())
            : undefined;
        this.logger.log('Function Deps: ' + deps.length);
        this.logger.log('hooksReference: ' + JSON.stringify(hooksReference));
        return {
            kind: types_1.RefactorKind.useCallback,
            func,
            deps,
            hooksReference
        };
    }
    getEditsForConvertUseCallback(info, file, textChangesContext) {
        const { deps, func, hooksReference } = info;
        const edits = this.typescript.textChanges.ChangeTracker.with(textChangesContext, changeTracker => {
            this.wrapIntoUseCallback(func, file, deps, hooksReference, changeTracker);
        });
        return {
            edits
        };
    }
    getEditsForConvertUseMemo(info, file, textChangesContext) {
        const { deps, expression, hooksReference } = info;
        const edits = this.typescript.textChanges.ChangeTracker.with(textChangesContext, changeTracker => {
            this.wrapIntoUseMemo(expression, file, deps, hooksReference, changeTracker);
        });
        return {
            edits
        };
    }
    wrapIntoUseCallback(func, file, deps, hooksReference, changeTracker) {
        const factory = this.typescript.factory;
        const [referenceExpression, postAction] = utils_1.createHooksReference(this.typescript, file, hooksReference, constants_1.Constants.UseCallback);
        const useCallbackCall = factory.createCallExpression(referenceExpression, undefined, [
            utils_1.functionExpressionLikeToExpression(this.typescript, func),
            factory.createArrayLiteralExpression(utils_1.dummyDeDuplicateDeps(deps).map(dep => utils_1.cloneDeep(this.typescript, dep)), false)
        ]);
        changeTracker.replaceNode(file, func, useCallbackCall);
        postAction === null || postAction === void 0 ? void 0 : postAction(changeTracker);
    }
    wrapIntoUseMemo(expression, file, deps, hooksReference, changeTracker) {
        const factory = this.typescript.factory;
        const [referenceExpression, postAction] = utils_1.createHooksReference(this.typescript, file, hooksReference, constants_1.Constants.UseMemo);
        const useMemoCall = factory.createCallExpression(referenceExpression, undefined, [
            factory.createArrowFunction(undefined, undefined, [], undefined, undefined, expression),
            factory.createArrayLiteralExpression(utils_1.dummyDeDuplicateDeps(deps).map(dep => utils_1.cloneDeep(this.typescript, dep)), false)
        ]);
        const finallyExpression = utils_1.wrapIntoJsxExpressionIfNeed(this.typescript, expression, useMemoCall);
        changeTracker.replaceNode(file, expression, finallyExpression);
        postAction === null || postAction === void 0 ? void 0 : postAction(changeTracker);
    }
    getRefactorContext(fileName) {
        if (!this.isValidJsxFlag()) {
            this.logger.log('Jsx options invalid');
            return undefined;
        }
        if (!this.typescript.fileExtensionIs(fileName, this.typescript.Extension.Tsx)) {
            this.logger.log('Not in tsx');
            return undefined;
        }
        const program = this.info.languageService.getProgram();
        if (!program) {
            this.logger.log('Cannot find program');
            return undefined;
        }
        const file = program.getSourceFile(fileName);
        if (!file) {
            this.logger.log('Cannot find source file');
            return undefined;
        }
        return {
            file,
            program
        };
    }
    getOutsideReferences(scope, additionalScope, file, checker, preferFullAccess = true) {
        const ts = this.typescript;
        const logger = this.logger;
        const references = [];
        const resolver = utils_1.createDepSymbolResolver(ts, scope, additionalScope, file, checker);
        visitor(scope);
        return references;
        function visitor(node) {
            if (ts.isTypeNode(node)) {
                return;
            }
            switch (node.kind) {
                case ts.SyntaxKind.Identifier: {
                    const identifier = node;
                    logger.log('Found Identifier: ' + identifier.text);
                    if (utils_1.shouldExpressionInDeps(ts, identifier, checker, resolver)) {
                        references.push(identifier);
                    }
                    return;
                }
                case ts.SyntaxKind.ElementAccessExpression:
                case ts.SyntaxKind.PropertyAccessExpression: {
                    const accessExpression = node;
                    logger.log('Found accessExpression: ' + accessExpression.getText());
                    if (preferFullAccess) {
                        if (utils_1.shouldExpressionInDeps(ts, accessExpression, checker, resolver)) {
                            references.push(accessExpression);
                            return;
                        }
                    }
                    if (ts.isPropertyAccessExpression(accessExpression)) {
                        visitor(accessExpression.expression);
                    }
                    else {
                        ts.forEachChild(accessExpression, visitor);
                    }
                    return;
                }
                default:
                    ts.forEachChild(node, visitor);
            }
        }
    }
    isValidJsxFlag() {
        const compilerOptions = this.info.languageServiceHost.getCompilationSettings();
        switch (compilerOptions.jsx) {
            case this.typescript.JsxEmit.React:
            case this.typescript.JsxEmit.ReactJSX:
            case this.typescript.JsxEmit.ReactJSXDev:
            case this.typescript.JsxEmit.ReactNative:
                return true;
            default:
                return false;
        }
    }
    isNewJsxTransformer() {
        const compilerOptions = this.info.languageServiceHost.getCompilationSettings();
        switch (compilerOptions.jsx) {
            case this.typescript.JsxEmit.ReactJSX:
            case this.typescript.JsxEmit.ReactJSXDev:
                return true;
            default:
                return false;
        }
    }
}
exports.CustomizedLanguageService = CustomizedLanguageService;
//# sourceMappingURL=service.js.map